def test_mccode_pooch_tags():
    from mccode_antlr.reader.registry import MCSTAS_REGISTRY, MCXTRACE_REGISTRY, LIBC_REGISTRY
    for reg in (MCSTAS_REGISTRY, MCXTRACE_REGISTRY, LIBC_REGISTRY):
        assert reg.version != "main"



def pre_mid_post(string, open_token, close_token):
    io = string.find(open_token)
    ic = string.rfind(close_token)
    if 0 < io < ic:
        return string[:io], string[io+1:ic], string[ic+1:]
    return None, string, None


class BaseWrapper:

    def __init__(self, hider: str = None, hidden: str = None, width: int = 80):
        self.hider = hider or 'hider'
        self.hidden = hidden or 'hidden'
        self.width = width

    @staticmethod
    def wrap(text: str, pad: str = None) -> list[str]:
        raise NotImplementedError

    @staticmethod
    def comment(name: str, content: str) -> str:
        raise NotImplementedError

    @staticmethod
    def start_block_comment(name: str) -> str:
        raise NotImplementedError

    @staticmethod
    def end_block_comment() -> str:
        raise NotImplementedError

    def block(self, name: str, content: str) -> str:
        raise NotImplementedError

    def line(self, name: str, items: list[str], sep: str = ' ') -> str:
        raise NotImplementedError

    def quoted_line(self, name: str, items: list[str], sep: str = ' ') -> str:
        raise NotImplementedError

    def lines(self, items: list[str]) -> str:
        raise NotImplementedError

    @staticmethod
    def metadata_group(name: str, mimetype: str, item: str, value: str) -> str:
        raise NotImplementedError

    @staticmethod
    def datatype(data_type: str) -> str:
        raise NotImplementedError

    @staticmethod
    def parameter(parameter: str) -> str:
        raise NotImplementedError

    @staticmethod
    def unit(unit: str) -> str:
        raise NotImplementedError

    @staticmethod
    def value(value: str) -> str:
        raise NotImplementedError

    @staticmethod
    def start_list(name: str) -> str:
        raise NotImplementedError

    @staticmethod
    def end_list(name: str) -> str:
        raise NotImplementedError

    @staticmethod
    def start_list_item() -> str:
        raise NotImplementedError

    @staticmethod
    def end_list_item() -> str:
        raise NotImplementedError

    @staticmethod
    def list_item(content: str) -> str:
        raise NotImplementedError

    @staticmethod
    def list_items(items: list[str]) -> str:
        raise NotImplementedError

    @staticmethod
    def escape(content: str) -> str:
        raise NotImplementedError

    @staticmethod
    def url(content: str) -> str:
        raise NotImplementedError

    @staticmethod
    def bold(content: str) -> str:
        raise NotImplementedError

    @staticmethod
    def emph(content: str) -> str:
        raise NotImplementedError

    @staticmethod
    def br() -> str:
        raise NotImplementedError

    @staticmethod
    def hide(content: str) -> str:
        raise NotImplementedError


class TextWrapper(BaseWrapper):

    def wrap(self, text: str, pad: str = None) -> list[str]:
        if pad is None:
            pad = ''
        return [self._wrap_line(line, self.width, pad) for line in text.splitlines(True)]

    def comment(self, name: str, content: str) -> str:
        pad = ' '*(len(name) + 1)
        lines = self.wrap(content, pad=pad)
        lines[0] = f'{name} {lines[0].strip()}'
        return '/*' + '\n* '.join(lines) + '\n*/'

    @staticmethod
    def start_block_comment(name: str) -> str:
        return f'/* {name}'

    @staticmethod
    def end_block_comment() -> str:
        return '*/'

    def block(self, name: str, content: str) -> str:
        return '\n'.join([name + '\n%{\n', content + '\n%}\n'])

    def line(self, name: str, items: list[str], sep: str = ' ') -> str:
        items = [item or 'None' for item in items]
        return '\n'.join(self.wrap(f'{name} {sep.join(items)}'))

    def quoted_line(self, name: str, items: list[str], sep: str = ' ') -> str:
        items = [item or 'None' for item in items]
        return '\n'.join(self.wrap(f'{name} "{sep.join(items)}"'))

    def lines(self, items: list[str]) -> str:
        return '\n'.join('\n'.join(self.wrap(line)) for line in items)

    @staticmethod
    def metadata_group(name: str, mimetype: str, item: str, value: str) -> str:
        return f'{name} "{mimetype}" {item} \n%{{\n{value}\n%}}\n'

    @staticmethod
    def datatype(data_type: str) -> str:
        return data_type

    @staticmethod
    def parameter(parameter: str) -> str:
        return f'{parameter}'

    @staticmethod
    def unit(unit: str) -> str:
        return f'/{unit}'

    @staticmethod
    def value(value: str) -> str:
        return f'{value}'

    @staticmethod
    def start_list(name: str) -> str:
        return name

    @staticmethod
    def end_list(name: str) -> str:
        return name

    @staticmethod
    def start_list_item() -> str:
        return ''

    @staticmethod
    def end_list_item() -> str:
        return ''

    @staticmethod
    def list_item(content: str) -> str:
        return content

    @staticmethod
    def list_items(items: list[str]) -> str:
        return '\n'.join([item for item in items])

    @staticmethod
    def escape(content: str) -> str:
        return content

    @staticmethod
    def url(content: str) -> str:
        return content

    @staticmethod
    def bold(content: str) -> str:
        return content

    @staticmethod
    def emph(content: str) -> str:
        return content

    @staticmethod
    def br() -> str:
        return ''

    @staticmethod
    def hide(content: str) -> str:
        return content

    @staticmethod
    def _wrap_line(line, width, pad) -> str:
        # This is not good enough. TODO Use a combined McCode _and_ C parser to identify reasonable line break points
        return line


class HTMLWrapper(BaseWrapper):


    @staticmethod
    def wrap(text: str, pad: str = None) -> list[str]:
        return [text]

    @staticmethod
    def comment(name: str, content: str) -> str:
        return f"<details><summary>{name}</summary>{content}</details>"

    @staticmethod
    def start_block_comment(name: str) -> str:
        return f'<details><summary>{name}</summary>'

    @staticmethod
    def end_block_comment() -> str:
        return '</details>'

    def block(self, name: str, content: str) -> str:
        return '<br>'.join([self.bold(name) + ' %{' + self.hide(f'<pre>{content}</pre><br>') + '%} '])

    def line(self, name: str, items: list[str], sep: str = ' ') -> str:
        return ''.join(self.wrap(f'<b>{name}</b> {sep.join(items)}')) + '<br>'

    def quoted_line(self, name: str, items: list[str], sep: str = ' ') -> str:
        return ''.join(self.wrap(f'<b>{name}</b> "{sep.join(items)}"')) + '<br>'

    def lines(self, items: list[str]) -> str:
        return '<br>'.join('<br>'.join(self.wrap(line)) for line in items)

    @staticmethod
    def metadata_group(name: str, mimetype: str, item: str, value: str) -> str:
        return f'<b>{name}</b> <code>"{mimetype}"</code>" <var>{item}</var> %{{<pre>{value}</pre>%}}'

    @staticmethod
    def datatype(data_type: str) -> str:
        return f'<code>{data_type}</code>'

    @staticmethod
    def parameter(parameter: str) -> str:
        return f'<var>{parameter}</var>'

    @staticmethod
    def unit(unit: str) -> str:
        return f'/{unit}'

    def value(self, value: str) -> str:
        return f'<code>{self.escape(str(value))}</code>'

    @staticmethod
    def start_list(name: str) -> str:
        return f'<ul> {name}'

    @staticmethod
    def end_list(name: str) -> str:
        return f'</ul> {name}'

    @staticmethod
    def start_list_item() -> str:
        return '<li>'

    @staticmethod
    def end_list_item() -> str:
        return '</li>'
    @staticmethod
    def list_item(content: str) -> str:
        return f'<li>{content}</li>'

    @staticmethod
    def list_items(items: list[str]) -> str:
        return ''.join([f'<li>{item}</li>' for item in items])

    @staticmethod
    def escape(content: str) -> str:
        from html import escape
        return escape(str(content))

    @staticmethod
    def url(content: str) -> str:
        return f'<a href="{content}">{content}</a>'

    @staticmethod
    def bold(content: str) -> str:
        return f'<b style="font-size:small">{content}</b>'

    @staticmethod
    def emph(content: str) -> str:
        return f'<em>{content}</em>'

    @staticmethod
    def br() -> str:
        return '<br>'

    def hide(self, content: str) -> str:
        from uuid import uuid4

        def span(c, tag, i, cont):
            return f'<span class="{c}" {tag}="{i}">{cont}</span>'

        if len(content.strip()):
            uuid = 'g-' + str(uuid4())  # a random UUID that _does not_ start with a 0
            return span(self.hider, 'data-id', uuid, span(self.hidden, 'id', uuid, content))

        return content



from dataclasses import dataclass
from typing import Union
from .expression import Expr


@dataclass
class InstrumentParameter:
    name: str
    unit: str
    value: Expr

    def to_file(self, output, wrapper):
        from .expression import DataType
        line = ''
        if self.value.is_str:
            line = 'string '
        elif self.value.is_vector and self.value.data_type == DataType.float:
            line = 'vector '
        elif self.value.data_type == DataType.int:
            line = 'int '
        line = wrapper.datatype(line) + wrapper.parameter(self.name)
        if self.unit is not None:
            line += wrapper.unit(self.unit)
        if self.value.has_value:
            line += "=" + wrapper.value(self.value)
        print(line, file=output)

    def to_string(self, wrapper):
        from io import StringIO
        output = StringIO()
        self.to_file(output, wrapper)
        return output.getvalue().strip()

    def __str__(self):
        from mccode_antlr.common import TextWrapper
        return self.to_string(TextWrapper())

    @staticmethod
    def parse(s: str):
        from antlr4 import InputStream
        from ..grammar import McInstr_parse
        from ..instr import InstrVisitor
        visitor = InstrVisitor(None, None)
        tree = McInstr_parse(InputStream(s), 'instrument_parameter')
        return visitor.getInstrument_parameter(tree)

    def copy(self):
        return InstrumentParameter(self.name, self.unit, self.value.copy())


@dataclass
class ComponentParameter:
    name: str
    value: Expr

    def to_string(self, wrapper):
        from io import StringIO
        output = StringIO()
        self.to_file(output, wrapper)
        return output.getvalue().strip()

    def __str__(self):
        from mccode_antlr.common import TextWrapper
        return self.to_string(TextWrapper())

    def to_file(self, output, wrapper):
        print(wrapper.parameter(self.name) + '=' + wrapper.value(self.value), file=output)

    def compatible_value(self, value):
        return self.value.compatible(value, id_ok=True)

    def __eq__(self, other):
        if isinstance(other, ComponentParameter):
            return self.name == other.name and self.value == other.value
        return self.value == other


def parameter_name_present(parameters: tuple[Union[InstrumentParameter, ComponentParameter], ...],
                           name: str) -> bool:
    return any(name == x.name for x in parameters)



from dataclasses import dataclass
from enum import Enum


class DataSource:
    class Type(Enum):
        Component = 1
        Instance = 2
        Instrument = 3

    def __init__(self, source_type: Type, name: str):
        self._type = source_type
        self._name = name

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, new_name: str):
        self._name = new_name

    def is_a(self, source_type: Type):
        return self._type == source_type

    @property
    def type_name(self):
        return self._type.name

    @staticmethod
    def from_type_name_and_name(type_name: str, name: str):
        return DataSource(DataSource.Type[type_name], name)


@dataclass
class MetaData:
    source: DataSource
    name: str
    mimetype: str
    value: str

    def to_file(self, output, wrapper):
        print(wrapper.metadata_group('METADATA', self.mimetype, self.name, self.value), file=output)

    @staticmethod
    def partial_from_tokens(source: DataSource, mimetype: str, name: str, value: str):
        if ' ' in name and (name[0] != '"' or name[-1] != '"'):
            # need to differentiate names with spaces from multiple names,
            # C runtime needs escaped quote characters for printing:
            name = '"' + name + '"'
        # remove the %{ and %} sentinels, if they're present
        if '%{' == value[:2] and '%}' == value[-2:]:
            value = value[2:-2]
        return MetaData(source, name, mimetype, value)

    @staticmethod
    def from_component_tokens(source: str, mimetype: str, name: str, value: str):
        return MetaData.partial_from_tokens(source=DataSource(DataSource.Type.Component, source),
                                            mimetype=mimetype, name=name, value=value)

    @staticmethod
    def from_instance_tokens(source: str, mimetype: str, name: str, value: str):
        return MetaData.partial_from_tokens(source=DataSource(DataSource.Type.Instance, source),
                                            mimetype=mimetype, name=name, value=value)

    @staticmethod
    def from_instrument_tokens(source: str, mimetype: str, name: str, value: str):
        return MetaData.partial_from_tokens(source=DataSource(DataSource.Type.Instrument, source),
                                            mimetype=mimetype, name=name, value=value)

    # output to metadata_table_struct initializer list
    def to_table_row(self):
        from .utilities import escape_str_for_c as c
        return ','.join(c(x) for x in (self.source.name, self.name, self.mimetype, self.value))

    def copy(self):
        return MetaData(self.source, self.name, self.mimetype, self.value)


from unittest import TestCase


class TestExpression(TestCase):
    def test_ObjectType(self):
        from mccode_antlr.common.expression import ObjectType as OT

        for ot in (OT.value, OT.initializer_list):
            self.assertFalse(ot.is_id)
            self.assertFalse(ot.is_parameter)
            self.assertFalse(ot.is_function)

        ot = OT.identifier
        self.assertTrue(ot.is_id)
        self.assertFalse(ot.is_parameter)
        self.assertFalse(ot.is_function)

        ot = OT.function
        self.assertFalse(ot.is_id)
        self.assertFalse(ot.is_parameter)
        self.assertTrue(ot.is_function)

        ot = OT.parameter
        self.assertFalse(ot.is_id)
        self.assertTrue(ot.is_parameter)
        self.assertFalse(ot.is_function)

        types = OT.value, OT.initializer_list, OT.identifier, OT.function, OT.parameter
        for ot in types:
            self.assertEqual(ot, ot)

        for i in range(len(types)):
            for j in range(len(types)):
                if i != j:
                    self.assertNotEqual(types[i], types[j])

    def test_ShapeType(self):
        from mccode_antlr.common.expression import ShapeType as ST
        ut = ST.unknown
        st = ST.scalar
        vt = ST.vector
        self.assertTrue(ut.compatible(st))
        self.assertTrue(ut.compatible(vt))
        self.assertTrue(ut.compatible(ut))
        self.assertTrue(st.compatible(ut))
        self.assertTrue(vt.compatible(ut))
        self.assertFalse(st.compatible(vt))
        self.assertFalse(vt.compatible(st))
        self.assertTrue(st.compatible(st))
        self.assertTrue(vt.compatible(vt))
        self.assertTrue(st.is_scalar)
        self.assertFalse(st.is_vector)
        self.assertTrue(vt.is_vector)
        self.assertFalse(vt.is_scalar)
        self.assertEqual(st.mccode_c_type, '')
        self.assertEqual(vt.mccode_c_type, '*')

    def test_DataType(self):
        from mccode_antlr.common.expression import DataType as DT
        u, f, i, s = DT.undefined, DT.float, DT.int, DT.str
        for x in (u, f, i, s):
            self.assertTrue(x.compatible(x))
            self.assertTrue(u.compatible(x))
            self.assertTrue(x.compatible(u))
        for x in (f, i):
            self.assertFalse(s.compatible(x))
            self.assertFalse(x.compatible(s))
        self.assertTrue(f.compatible(i))
        self.assertTrue(i.compatible(f))

        for x in (u, f, i, s):
            self.assertEqual(x + x, x)
            self.assertEqual(x - x, x)
            self.assertEqual(x * x, x)

        for x in (f, i, s):
            self.assertEqual(x + u, x)
            self.assertEqual(x - u, x)
            self.assertEqual(x * u, x)
            self.assertEqual(u + x, x)
            self.assertEqual(u - x, x)
            self.assertEqual(u * x, x)

        for x in (u, f, i):
            self.assertEqual(x // x, i)
            self.assertEqual(x / x, f)
            self.assertEqual(x / u, f)
            self.assertEqual(u / x, f)

        for x in (u, f, i):
            self.assertEqual(x + s, s)
            self.assertEqual(x - s, s)
            self.assertEqual(x * s, s)
            self.assertEqual(s + x, s)
            self.assertEqual(s - x, s)
            self.assertEqual(s * x, s)
            self.assertRaises(RuntimeError, lambda: x / s)
            self.assertRaises(RuntimeError, lambda: s / x)

        self.assertEqual(i + f, f)
        self.assertEqual(i - f, f)
        self.assertEqual(i * f, f)
        self.assertEqual(i / f, f)

        self.assertEqual(i.mccode_c_type, 'int')
        self.assertEqual(f.mccode_c_type, 'double')
        self.assertEqual(s.mccode_c_type, 'char *')

    def test_Value(self):
        from mccode_antlr.common.expression import Value, ObjectType, ShapeType
        float_value = Value.float(1)
        int_value = Value.int(1)
        str_value = Value.str('1')
        id_value = Value.id('one')
        array_value = Value.array([1])
        function_value = Value.function('ones')

        for x in (float_value, int_value, str_value, id_value, array_value, function_value):
            self.assertFalse(x.is_op)
            self.assertFalse(x.is_zero)

        self.assertEqual(float_value.object_type, ObjectType.value)
        self.assertEqual(int_value.object_type, ObjectType.value)
        self.assertEqual(str_value.object_type, ObjectType.value)
        self.assertEqual(id_value.object_type, ObjectType.identifier)
        self.assertEqual(array_value.object_type, ObjectType.value)
        self.assertEqual(function_value.object_type, ObjectType.function)

        for x, y in ((float_value, ShapeType.scalar), (int_value, ShapeType.scalar), (str_value, ShapeType.scalar),
                     (id_value, ShapeType.scalar), (array_value, ShapeType.vector), (function_value, ShapeType.scalar)):
            self.assertEqual(x.shape_type, y)

        for x in (float_value, int_value, id_value, array_value, function_value):
            self.assertFalse(x.is_str)
        self.assertTrue(str_value.is_str)

        for x in (float_value, int_value, str_value, array_value, function_value):
            self.assertTrue(x.is_value(x._value))

        # Identifiers are not values:
        self.assertTrue(id_value.is_id)
        self.assertFalse(id_value.is_value(id_value._value))

        for x in (float_value, int_value, str_value, id_value, function_value):
            self.assertTrue(x.is_scalar)
            self.assertFalse(x.is_vector)
        self.assertFalse(array_value.is_scalar)
        self.assertTrue(array_value.is_vector)

    def test_null_vector_Value(self):
        from mccode_antlr.common.expression import Value, ObjectType, ShapeType
        from loguru import logger
        array = Value.array("NULL")
        # an Array is compatible with a raw string (nor a Value.best(str) which produces an identifier)
        self.assertTrue(array.compatible("identifier"))
        # but not a quoted string
        self.assertFalse(array.compatible('"Not an identifier"'))

    def _numeric_Value_checks(self, one, two):
        self.assertEqual(one, one)
        self.assertEqual(one + one, two)
        self.assertEqual(two - one, one)
        self.assertEqual(one * one, one)
        self.assertEqual(two / two, one)
        self.assertEqual(one - two, -one)
        self.assertEqual(abs(one - two), one)
        self.assertTrue(one < two)
        self.assertTrue(one <= two)
        self.assertTrue(two > one)
        self.assertTrue(two >= one)

    def test_float_Value(self):
        from mccode_antlr.common.expression import Value
        one = Value.float(1)
        two = Value.float(2)
        self._numeric_Value_checks(one, two)
        self.assertEqual(one.mccode_c_type, 'double')
        self.assertEqual(one.mccode_c_type_name, 'instr_type_double')

    def test_int_Value(self):
        from mccode_antlr.common.expression import Value
        one = Value.int(1)
        two = Value.int(2)
        self._numeric_Value_checks(one, two)
        self.assertEqual(one.mccode_c_type, 'int')
        self.assertEqual(one.mccode_c_type_name, 'instr_type_int')

    def test_str_Value(self):
        from mccode_antlr.common.expression import Value, BinaryOp
        one = Value.str('one')
        two = Value.str('two')
        self.assertEqual(one, one)
        self.assertEqual(one + one, BinaryOp('+', one, one))
        self.assertEqual(two - one, BinaryOp('-', two, one))
        self.assertEqual(one * one, BinaryOp('*', one, one))
        self.assertRaises(RuntimeError, lambda: two / one)
        self.assertEqual(one - two, BinaryOp('-', one, two))
        self.assertEqual(one.mccode_c_type, 'char *')
        self.assertEqual(one.mccode_c_type_name, 'instr_type_string')

    def test_vector_Value(self):
        from mccode_antlr.common.expression import Value, DataType
        first = Value.array([1, 2, 3], DataType.float)
        second = Value.array([4, 5, 6], DataType.int)

        self.assertTrue(first.is_vector)
        self.assertEqual(first.shape_type, second.shape_type)
        self.assertEqual(first.vector_len, second.vector_len)
        self.assertNotEqual(first.data_type, second.data_type)
        # Addition concatenates the arrays and promotes the datatype
        self.assertEqual(first + second, Value.array([1, 2, 3, 4, 5, 6], DataType.float))

    def test_id_Value(self):
        from mccode_antlr.common.expression import Value, BinaryOp
        id = Value.id('some_parameter')

        self.assertFalse(id.is_str)
        self.assertTrue(id.is_id)
        self.assertEqual(id + Value.int(1), BinaryOp('+', id, Value.int(1)))

    def test_parameter_Value(self):
        from mccode_antlr.common.expression import Value, ObjectType
        par = Value('instrument_parameter', object_type=ObjectType.parameter)
        self.assertFalse(par.is_str)
        # self.assertFalse(par.is_id)  # changed as of 2023-10-16 -- a parameter is an identifier
        self.assertTrue(par.is_parameter)
        # Verify that the whole reason for this object type existing works (inserting its name into macros)
        self.assertEqual(f'{par:p}', "_instrument_var._parameters.instrument_parameter")
        # The original implementation broke instrument printing, so str(par) should always be its bare name
        self.assertEqual(str(par), "instrument_parameter")
        self.assertEqual(f'{par}', 'instrument_parameter')
        # And, just in case it's useful, we can specify the prefix:
        self.assertEqual(f'{par:prefix:this_is_cool_}', "this_is_cool_instrument_parameter")

    def test_UnaryOp(self):
        from mccode_antlr.common.expression import Value, UnaryOp
        val = Value.int(1)
        # This one is not equal because the unary minus on a numeric value is pushed inside directly
        self.assertNotEqual(-val, UnaryOp('-', val))
        # But negating the explicit unary minus operation pops-out its contents
        self.assertEqual(-UnaryOp('-', val), val)
        self.assertEqual(round(val), val)

        val = Value.int(-1)
        self.assertNotEqual(-val, UnaryOp('-', val))
        self.assertEqual(-UnaryOp('-', val), val)
        # again, unary absolute value is pushed onto the contained integer directly
        self.assertNotEqual(abs(val), UnaryOp('abs', val))
        self.assertEqual(abs(val), Value.int(1))
        # abs(abs(x)) == abs(x) [even if we dont know anything about x]
        self.assertEqual(abs(UnaryOp('abs', val)), UnaryOp('abs', val))

        val = Value.str('string')
        # FIXME consider changing this behaviour, since negating a string does not make sense
        self.assertEqual(-val, UnaryOp('-', val))
        self.assertEqual(-UnaryOp('-', val), val)
        # Nor does the absolute value of a string ... but other (non Python) unary operations make sense
        self.assertEqual(abs(val), UnaryOp('abs', val))

        val = Value.id('identifier')
        self.assertEqual(-val, UnaryOp('-', val))
        self.assertEqual(abs(val), UnaryOp('abs', val))
        self.assertEqual(round(val), UnaryOp('round', val))

        val = Value.float(1.)
        # Again, some unary operations get pushed directly onto the underlying value
        self.assertEqual(round(val), val)
        # But if we make an explicit wrapped UnaryOp they do not
        self.assertNotEqual(round(UnaryOp('abs', val)), UnaryOp('abs', val))
        self.assertNotEqual(round(UnaryOp('abs', val)), val)

        not_val = UnaryOp('__not__', Value.id('val'))
        self.assertEqual(str(not_val), '!val')
        not_val.style = 'Python'
        self.assertEqual(str(not_val), 'not val')

    def test_numeric_BinaryOp(self):
        from mccode_antlr.common.expression import Value, BinaryOp
        f = [Value.float(x) for x in range(3)]
        i = [Value.float(x) for x in range(3)]
        # simple operations on numeric values are applied directly, which does not produce a BinaryOp object
        for a, b in ((f[1], f[2]), (i[1], i[2]), (f[1], i[2]), (i[1], f[2])):
            self.assertNotEqual(a + b, BinaryOp('+', a, b))
            self.assertNotEqual(a - b, BinaryOp('-', a, b))
            self.assertNotEqual(a * b, BinaryOp('*', a, b))
            self.assertNotEqual(a / b, BinaryOp('/', a, b))

        one = Value.id('one')
        two = Value.id('two')
        # simple operations on non-numeric values are not applied directly, which does produce a BinaryOp object
        for a, b in ((one, two), (one, i[2]), (one, f[2])):
            self.assertEqual(a + b, BinaryOp('+', a, b))
            self.assertEqual(a - b, BinaryOp('-', a, b))
            self.assertEqual(a * b, BinaryOp('*', a, b))
            self.assertEqual(a / b, BinaryOp('/', a, b))
        # except in the special case where we multiply or divide by a numeric 1
        for a, b in ((f[1], two), (i[1], two)):
            self.assertEqual(a + b, BinaryOp('+', a, b))
            self.assertEqual(a - b, BinaryOp('-', a, b))
            self.assertEqual(a * b, b)
            self.assertEqual(a / b, BinaryOp('/', a, b))
        for a, b in ((one, i[1]), (two, f[1])):
            self.assertEqual(a + b, BinaryOp('+', a, b))
            self.assertEqual(a - b, BinaryOp('-', a, b))
            self.assertEqual(a * b, a)
            self.assertEqual(a / b, a)
        # Or multiply by numeric zero
        self.assertTrue((one * f[0]).is_zero)
        self.assertTrue((two * i[0]).is_zero)
        self.assertTrue((i[0] / one).is_zero)
        self.assertTrue((f[0] / two).is_zero)

    def _style_tests(self, op, c_style, python_style):
        self.assertEqual(str(op), c_style)
        op.style = 'anything which is not just "C"'
        self.assertEqual(str(op), python_style)

    def test_identifier_BinaryOp(self):
        from mccode_antlr.common.expression import Value, BinaryOp
        one, two = [Value.id(x) for x in ('one', 'two')]
        # BinaryOp allows '__call__' with any value types, but 'left' _should_ have a object_type of ObjectType.function
        self.assertEqual(str(BinaryOp('__call__', one, two)), 'one(two)')
        # Other special binary operations have different representations in C vs Python
        for op, c, py in (('struct_access', 'one.two', 'getattr(one, "two")'),
                          ('pointer_access', 'one->two', 'getattr(one, "two")'),
                          ('pow', 'one^two', 'one**two'),
                          ('or', 'one || two', 'one or two'),
                          ('and', 'one && two', 'one and two')):
            self._style_tests(BinaryOp(f'__{op}__', one, two), c, py)

        self.assertEqual(str(BinaryOp('any_function', one, two)), 'any_function(one, two)')

    def test_identifier_TrinaryOp(self):
        from mccode_antlr.common.expression import Value, TrinaryOp
        test, one, two = [Value.id(x) for x in ('test', 'one', 'two')]
        self._style_tests(TrinaryOp('__trinary__', test, one, two), 'test ? one : two', 'one if test else two')

    def test_simple_Expr(self):
        pass

    def test_parse_Expr(self):
        from mccode_antlr.common.expression import Expr, UnaryOp, BinaryOp, TrinaryOp, Value, ObjectType

        self.assertEqual(Expr.parse('1'), Value.int(1))
        self.assertEqual(Expr.parse('1.'), Value.float(1))
        self.assertEqual(Expr.parse('0'), Value.int(0))
        self.assertEqual(Expr.parse('0.'), Value.float(0))
        self.assertEqual(Expr.parse('.0'), Value.float(0))

        self.assertEqual(Expr.parse('1+1'), Value.int(2))
        self.assertEqual(Expr.parse('1.+1.'), Value.float(2))
        self.assertEqual(Expr.parse('1+1.0'), Value.float(2))
        self.assertEqual(Expr.parse('1.0+1'), Value.float(2))

        self.assertEqual(Expr.parse('1-1'), Value.int(0))
        self.assertEqual(Expr.parse('1.-1.'), Value.float(0))
        self.assertEqual(Expr.parse('1-1.0'), Value.float(0))
        self.assertEqual(Expr.parse('1.0-1'), Value.float(0))

        self.assertEqual(Expr.parse('1*1'), Value.int(1))
        self.assertEqual(Expr.parse('1.*1.'), Value.float(1))
        self.assertEqual(Expr.parse('1*1.0'), Value.float(1))
        self.assertEqual(Expr.parse('1.0*1'), Value.float(1))

        self.assertEqual(Expr.parse('1/1'), Value.int(1))
        self.assertEqual(Expr.parse('1./1.'), Value.float(1))
        self.assertEqual(Expr.parse('1/1.0'), Value.float(1))
        self.assertEqual(Expr.parse('1.0/1'), Value.float(1))

        self.assertEqual(Expr.parse('"blah"'), Value.str('"blah"'))
        self.assertEqual(Expr.parse('blah'), Value.id('blah'))

        sin_minus_pi_x_over_2 = Expr.parse('sin( -PI * x / 2.   )')
        pi = Expr.id('PI')
        x = Expr.id('x')
        sin = Value('sin', object_type=ObjectType.function)
        # These two expressions should be identical, but details of the Lexer/Parser pick the second
        equiv_expr = BinaryOp('__call__', sin, BinaryOp('/', BinaryOp('*', UnaryOp('-', pi), x), Value.float(2)))
        expr = BinaryOp('__call__', sin,
                        UnaryOp('-', BinaryOp('/', BinaryOp('*', pi, x), Value.float(2)))
                        )

        self.assertEqual(sin_minus_pi_x_over_2, expr)
        self.assertEqual(str(sin_minus_pi_x_over_2), str(expr))
        self.assertEqual(str(sin_minus_pi_x_over_2), str(equiv_expr))

        # One reason that Expr has a list of expressions:
        atan2_y_x = Expr.parse('arctan2(y, x)')
        atan2 = Value('arctan2', object_type=ObjectType.function)
        y = Expr.id('y')
        expr = BinaryOp('__call__', atan2, Expr([y, x]))
        self.assertEqual(atan2_y_x, expr)

    def test_instrument_parameter(self):
        from antlr4 import InputStream
        from mccode_antlr.grammar import McInstr_parse
        from mccode_antlr.instr import InstrVisitor
        from mccode_antlr.reader import MCSTAS_REGISTRY, Reader
        from mccode_antlr.common.expression import Value, ObjectType
        instr_source = """
        DEFINE INSTRUMENT blah(int par=0)
        TRACE
        COMPONENT origin = Progress_bar() AT (0, 0, 0) RELATIVE ABSOLUTE
        COMPONENT source = Source_gen(T1=413.5, I1=10.22e12, T2=145.8, I2=3.44e13, T3=40.1, I3=2.78e13, dist=2,
                                      radius=0.06, focus_xw=0.1, focus_yh=0.1, lambda0=1.0, dlambda=0.5)
                           AT (0, 0, 0) RELATIVE PREVIOUS
        COMPONENT monitor = PSD_monitor(nx=par, ny=par, xwidth=0.1, yheight=0.1) AT (0, 0, 10.1) RELATIVE PREVIOUS
        END
        """
        tree = McInstr_parse(InputStream(instr_source), 'prog')
        visitor = InstrVisitor(Reader(registries=[MCSTAS_REGISTRY, ]), None)
        # Parse the instrument definition and return an Instr object
        instr = visitor.visitProg(tree)
        nx = instr.components[-1].get_parameter('nx')
        self.assertTrue(nx.value.is_parameter)
        self.assertEqual(nx.value, Value('par', object_type=ObjectType.parameter))

    def test_simplify(self):
        from mccode_antlr.common.expression import Expr, Value
        is_two = Expr.parse('(2+4)/(1+2)')
        self.assertTrue(isinstance(is_two, Expr))
        is_two = is_two.simplify()
        # Simplifying an expression yields an expression
        self.assertTrue(isinstance(is_two, Expr))
        # Which is equal to the expected result of 2
        self.assertEqual(is_two, Expr.int(2))
        # For it to be constant it can only have one sub expression part:
        self.assertEqual(len(is_two.expr), 1)
        # Which must be a Value
        self.assertTrue(isinstance(is_two.expr[0], Value))
        # That is not an identifier
        self.assertFalse(is_two.expr[0].is_id)
        # Finally, the real test -- successfully simplifying an Expr produces a constant Expr
        self.assertTrue(is_two.is_constant)

    def test_evaluate(self):
        from mccode_antlr.common.expression import Expr, Value
        known = {'bw1phase': Expr.float(0)}
        phase = Expr.parse('bw1phase / 2')
        self.assertTrue(isinstance(phase, Expr))
        phase = phase.evaluate(known)
        # Evaluating an expression yields an expression
        self.assertTrue(isinstance(phase, Expr))
        # Which is equal to the expected result of 0
        self.assertEqual(phase, Expr.float(0))
        # For it to be constant it can only have one sub expression part:
        self.assertEqual(len(phase.expr), 1)
        # Which must be a Value
        self.assertTrue(isinstance(phase.expr[0], Value))
        # That is not an identifier
        self.assertFalse(phase.expr[0].is_id)
        # Finally, the real test -- successfully evaluating (and simplifying) an Expr produces a constant Expr
        self.assertTrue(phase.is_constant)

    def test_depends_on(self):
        from mccode_antlr.common.expression import Expr
        phase = Expr.parse('bw1phase / 2')
        self.assertTrue(phase.depends_on('bw1phase'))

        expr = Expr.parse('floor(1.4445)/sin(pi * angle / 180.)')
        for name in ('angle', 'pi'):
            self.assertTrue(expr.depends_on(name))
        for func_name in ('floor', 'sin'):
            self.assertFalse(expr.depends_on(func_name))
        for not_name in ('1.4445', '180.', 1.445, 180.0):
            self.assertFalse(expr.depends_on(not_name))

    def test_numeric_operations(self):
        from mccode_antlr.common.expression import Expr, DataType
        x = Expr.float(1.)
        for res in (1 + x, x + 1, 2 * x, x / 0.5, 2 / x):
            self.assertTrue(isinstance(res, Expr))
            self.assertTrue(res.data_type == DataType.float)
            self.assertEqual(res, Expr.float(2))
            self.assertEqual(res, 2)

    def test_numeric_expressions(self):
        from mccode_antlr.common.expression import Expr
        for n_slits in (2, Expr.int(2)):
            for theta_0 in (Expr.float(100), Expr.id('variable')):
                delta = theta_0 / 2.0
                edges = [y * 360.0 / n_slits + x for y in range(int(n_slits)) for x in (-delta, delta)]
                self.assertEqual(len(edges), 4)
                for edge in edges:
                    self.assertTrue(isinstance(edge, Expr))

    def test_is_vector(self):
        from mccode_antlr.common.expression import Expr, Value, ObjectType, ShapeType, DataType, BinaryOp
        self.assertTrue(Expr.array([1, 2, 3]).is_vector)
        self.assertFalse(Expr.int(1).is_vector)

        # This is needed for setting, e.g., component parameters from declared _vector_ variables
        ex = Expr.id('ex')
        values = Value('values', object_type=ObjectType.identifier, shape_type=ShapeType.vector, data_type=DataType.float)
        expr = -ex / BinaryOp('__getitem__', values, Expr.int(0))
        self.assertEqual(f'{expr}', '-ex / values[0]')

        self.assertFalse(expr.is_vector) # because values[0] is a scalar


def version():
    import sys
    if sys.version_info[0] == 3 and sys.version_info[1] < 8:
        import importlib_metadata
    else:
        import importlib.metadata as importlib_metadata
    try:
        return importlib_metadata.version("mccode_antlr")
    except importlib_metadata.PackageNotFoundError:
        return "dev"

